
#ifndef __KERNEL__
 #define __KERNEL__
#endif

#ifndef MODULE
 #define MODULE
#endif

#include <linux/config.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/fcntl.h>
#include <linux/ioctl.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <asm/system.h>
#include <asm/segment.h>
#include <asm/io.h>
#include <linux/ioport.h>
#include <linux/init.h>

#include "disa8255multi.h"
#include "sysdep.h"   

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
#include <asm/uaccess.h>
#endif

/* I defined some macro for device_open function */
#define TYPE(dev) (MINOR(dev) >> 4) 
#define NUM(dev) (MINOR(dev) & 0x0f)

#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c) ((a)*65536+(b)*256+(c))
#endif

static int cards;
static char rev[2];
static int direction[6];
static int counterbcd[3];
static int countermode[3];
int Base_Addr=-1;
MODULE_PARM(Base_Addr,"i");
EXPORT_NO_SYMBOLS;

/*---------------------------------------------------------------------*/
int disa8255multi_open(struct inode *inoder,struct file *flip)
{
/* int type= TYPE(inoder->i_rdev);*/

 int num=NUM(inoder->i_rdev);
 if(num>=cards) return -ENODEV;

 MOD_INC_USE_COUNT;
 return 0;
}
/*---------------------------------------------------------------------*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
static int disa8255multi_release(struct inode *inoder,struct file *flip)
#else
static void disa8255multi_release(struct inode *inoder,struct file *flip)
#endif
{
  MOD_DEC_USE_COUNT;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
  return 0;
#endif    
}
/*---------------------------------------------------------------------*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
static ssize_t disa8255multi_read(struct file *flip,char *buf,size_t count,loff_t *ptr)
#else
static int disa8255multi_read(struct inode *inode,struct file *flip, char *buf,count_t count)
#endif
{
 int countt=0;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
  int num=MINOR(flip->f_dentry->d_inode->i_rdev);
  if(ptr!=&flip->f_pos)  /* can't seek on this device */
   return -ESPIPE;
#else
  int num=NUM(inode->i_rdev);
#endif 
 
 if(num>=cards)
   return -EINVAL;
 buf[0]=readb(Base_Addr);
 countt=1;
           
 return countt;
}
/*---------------------------------------------------------------------*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
static ssize_t disa8255multi_write(struct file *flip,const char *buf,size_t count,loff_t *ptr)
#else
static int disa8255multi_write(struct inde *inode,struct file *flip,char *buf,count_t count)
#endif
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)	
 if(ptr!=&flip->f_pos)  /* can't seek on this device */
  return -ESPIPE;
#endif
 
 return 0;       /* do nothing */
}
/*---------------------------------------------------------------------*/
int Myread(int num,char *buf,int count)
{
 if((buf[0]>=0)&&(buf[0]<10))
   buf[2]=inb(Base_Addr+(int)buf[1]*4+(int)buf[0]);
 else
  {    
   if(buf[0]==10)                /* control word */ 
    buf[2]=inb(Base_Addr+(int)buf[1]*4+3);
  }
  return 3;
}
/*---------------------------------------------------------------------*/
void Mywrite(int num,char *buf,int count)
{
 if((buf[0]>=0)&&(buf[0]<10))
   outb(buf[2],Base_Addr+(int)buf[1]*4+(int)buf[0]);
 else
  {    
   if(buf[0]==10)                /* control word */ 
    outb(buf[2],Base_Addr+(int)buf[1]*4+3);
  }
}
/*---------------------------------------------------------------------*/
int disa8255multi_ioctl(
    struct inode *inode,
    struct file *file,
    unsigned int ioctl_num,/* The number of the ioctl */
    unsigned long ioctl_param) /* The parameter to it */
{
  char param[4];
  char *temp=(char *)ioctl_param;
  char ch1=0;
  int data;
  int num=NUM(inode->i_rdev);
  switch (ioctl_num) 
   {

    case DISA8255MULTI_SET_REGISTER:
      ch1=(char)((int)temp);
      if(ch1<30)
       {
        rev[0]=ch1/30;
        rev[1]=ch1;
        return 0;
       }
       
      if(ch1>29)
       {
        rev[0]=10;
        rev[1]=ch1-30;
       } 
      break;

    case DISA8255MULTI_WRITE_PORT:
      param[0]=rev[0];
      param[1]=rev[1];
      param[2]=(char)((int)temp);
      Mywrite(num,param,4);
      break;

    case DISA8255MULTI_READ_PORT:
      param[0]=rev[0];
      param[1]=rev[1];
      Myread(num,param,4);
      put_user(param[2],temp);
      break;

    case DISA8255MULTI_READ_COUNTER:
      param[0]=rev[0];
      param[1]=rev[1];
      Myread(num,param,4);
      put_user(param[2],temp);
      break; 
      
    case DISA8255MULTI_WRITE_COUNTER:
      data=(int)ioctl_param;
      param[0]=rev[0];
      param[1]=rev[1];
      param[2]=(char)data;
      param[3]=(char)(data>>8);
      Mywrite(num,param,4);
      break;

    case DISA8255MULTI_SET_CONTROLWORD:
      param[0]=10;
      param[1]=rev[1];
      param[2]=(char)((int)temp);
      Mywrite(num,param,4);
      break;      
            
    case DISA8255MULTI_GET_CONTROLWORD:
      param[0]=10;
      param[1]=rev[1];
      Myread(num,param,4);
      put_user(param[2],temp);
      break;
     default:
             break;
  }

  return 0;
}
/*---------------------------------------------------------------------*/
/*---------------------------------------------------------------------*/
static struct file_operations disa8255multi_fops={
NULL,  /* disa8255multi_lseek, */           
&disa8255multi_read,
&disa8255multi_write,
NULL,  /* disa8255multi_readdir */
NULL,  /* disa8255multi_select */
&disa8255multi_ioctl ,
NULL,  /* disa8255multi_mmap */
&disa8255multi_open,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
NULL,   /*  flush*/
#endif
&disa8255multi_release,
};
/*---------------------------------------------------------------------*/
/*---------------------------------------------------------------------*/
/*---------------------------------------------------------------------*/
/*---------------------------------------------------------------------*/
MODULE_AUTHOR("Delta");
MODULE_DESCRIPTION("Decision Multi ISA 8255 Driver");
/*---------------------------------------------------------------------*/
/*---------------------------------------------------------------------*/

int init_module(void)
{
 int index=0;
 static int dev_count=0;
 int i; 

 dev_count++;
 if(dev_count>1)
  {
   printk(KERN_ERR"This drive only can open one time.\n");
   return -ENODEV;
  }

 cards=0;
/*--------------*/ 
if(Base_Addr==-1)
 {
 printk(KERN_ERR "Multi ISA 8255: I need io address!!\n");
 return -EINVAL;
 }
   printk("Multi ISA 8255:I use 0x%x to be my IO address\n",Base_Addr); 
   if(check_region(Base_Addr,REGIONLENGTH)>=0)
    request_region(Base_Addr,REGIONLENGTH,DEVICE_NAME);
   
   for(i=0;i<6;i++)
    direction[i]=0;  // default is output
   for(i=0;i<3;i++)
    {
     countermode[i]=3;
     counterbcd[i]=0;
    } 
   index++;
/*  } */

  if(index<1)
   {
    printk("\nI can't find any Decision Multi ISA 8255 card here!!\n");
    return -ENODEV;
   }
 cards=index;
 index=register_chrdev(disa8255multi_major,DEVICE_NAME,&disa8255multi_fops);
 if(index<0)
  {
  release_region(Base_Addr,REGIONLENGTH);
  return index;
  }
 rev[0]=0;
 rev[1]=0;
 printk("Multi ISA 8255:Register Successful\n");  
 return 0;

}
/*---------------------------------------------------------------------*/
void cleanup_module(void)
{
  release_region(Base_Addr,REGIONLENGTH); 
  unregister_chrdev(disa8255multi_major,DEVICE_NAME);
  printk("Multi ISA 8255:Unregister Successful!!\n");  
}
/*---------------------------------------------------------------------*/

